"""
Type annotations
================

:mod:`dawiq.typing` provides special type annotations.
"""

from typing import Protocol, Dict, Any
from .qt_compat import QtCore


__all__ = [
    "DataclassProtocol",
    "FieldWidgetProtocol",
]


class DataclassProtocol(Protocol):
    """Type annotation for dataclass type object."""

    # https://stackoverflow.com/a/55240861/11501976
    __dataclass_fields__: Dict


class FieldWidgetProtocol(Protocol):
    """Type annotation for field widget object."""

    dataValueChanged: QtCore.Signal

    def dataValue(self) -> Any:
        """
        Data value that the widget represents.

        This is the API for the delegate to get the data from the widget.
        For example, in :class:`BoolCheckBox <dawiq.fieldwidgets.BoolCheckBox>`
        this method converts :class:`bool` to :obj:`Qt.CheckState` and sets the
        check state.

        If the data value is :obj:`dawiq.MISSING`, it indicates that the field is
        empty and delegate should handle it specially.

        When the data value is changed by user input, :attr:`dataValueChanged`
        signal must emit the new value.

        """
        ...

    def setDataValue(self, value: Any):
        """
        Set the data value to the widget.

        This method is the API for the delegate to set the data to the widget.
        For example, in :class:`BoolCheckBox <dawiq.fieldwidgets.BoolCheckBox>`
        this method converts the check state to :class:`bool` and returns.

        This method must specially treat :obj:`dawiq.MISSING` as empty data by
        clearing the widget.

        This method MUST NOT emit :attr:`dataValueChanged` signal, as doing so
        can cause infinite loop.
        """
        ...

    def fieldName(self) -> str:
        """
        Name of the field.

        It is recommended to make this name visible in the widget without
        affecting the data value. Placeholder text of the line edit or title of
        the group box are good examples.

        """
        ...

    def setFieldName(self, name: str):
        """Set the name of the field."""
        ...



from dawiq import (
    DataWidget,
    type2Widget,
    dataclass2Widget,
    BoolCheckBox,
    IntLineEdit,
    FloatLineEdit,
    StrLineEdit,
    EnumComboBox,
    TupleGroupBox,
    MISSING,
)
from dawiq.qt_compat import QtCore
import dataclasses
from enum import Enum
from typing import Optional, Tuple
import pytest


def test_DataWidget_addWidget(qtbot):
    datawidget = DataWidget()
    assert datawidget.count() == 0

    w0 = BoolCheckBox()
    w0.setFieldName("w0")
    datawidget.addWidget(w0)
    assert datawidget.count() == 1
    assert datawidget.widget(0) is w0
    assert datawidget.widget(1) is None

    w1 = IntLineEdit()
    w1.setFieldName("w0")
    with pytest.raises(KeyError):
        datawidget.addWidget(w1)

    w1.setFieldName("w1")
    datawidget.addWidget(w1)
    assert datawidget.count() == 2
    assert datawidget.widget(0) is w0
    assert datawidget.widget(1) is w1

    # test that signals are connected
    with qtbot.waitSignal(datawidget.dataValueChanged):
        w0.click()
    with qtbot.waitSignal(datawidget.dataValueChanged):
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)


def test_DataWidget_insertWidget(qtbot):
    datawidget = DataWidget()
    assert datawidget.count() == 0

    w0 = BoolCheckBox()
    w0.setFieldName("w0")
    datawidget.insertWidget(0, w0)
    assert datawidget.count() == 1
    assert datawidget.widget(0) is w0
    assert datawidget.widget(1) is None

    w1 = IntLineEdit()
    w1.setFieldName("w0")
    with pytest.raises(KeyError):
        datawidget.insertWidget(0, w1)

    w1.setFieldName("w1")
    datawidget.insertWidget(0, w1)
    assert datawidget.count() == 2
    assert datawidget.widget(0) is w1
    assert datawidget.widget(1) is w0

    # test that signals are disconnected
    with qtbot.waitSignal(datawidget.dataValueChanged):
        w0.click()
    with qtbot.waitSignal(datawidget.dataValueChanged):
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)


def test_DataWidget_removeWidget(qtbot):
    datawidget = DataWidget()
    w0 = BoolCheckBox()
    w0.setFieldName("w0")
    w1 = IntLineEdit()
    w1.setFieldName("w1")

    datawidget.addWidget(w0)
    assert datawidget.count() == 1

    datawidget.removeWidget(w1)
    assert datawidget.count() == 1

    datawidget.removeWidget(w0)
    assert datawidget.count() == 0

    # test that signals are disconnected
    with qtbot.assertNotEmitted(datawidget.dataValueChanged):
        w0.click()


def test_DataWidget_dataValue(qtbot):
    @dataclasses.dataclass
    class Cls1:
        x: int

    @dataclasses.dataclass
    class Cls2:
        a: int
        b: Cls1

    dataWidget = dataclass2Widget(Cls2)
    assert dataWidget.dataValue() == dict(a=MISSING, b=dict(x=MISSING))

    dataWidget.widget(0).setText("1")
    dataWidget.widget(1).widget(0).setText("2")
    assert dataWidget.dataValue() == dict(a=1, b=dict(x=2))


def test_DataWidget_setDataValue(qtbot):
    @dataclasses.dataclass
    class Cls1:
        x: bool

    @dataclasses.dataclass
    class Cls2:
        a: int
        b: Cls1

    dataWidget = dataclass2Widget(Cls2)

    class Counter:
        def __init__(self):
            self.i = 0

        def count(self):
            self.i += 1

    counter = Counter()
    dataWidget.dataValueChanged.connect(counter.count)
    dval = dict(a=1, b=dict(x=True))

    with qtbot.assertNotEmitted(dataWidget.dataValueChanged):
        dataWidget.setDataValue(dval)
    assert dataWidget.dataValue() == dval
    assert counter.i == 0

    with qtbot.assertNotEmitted(dataWidget.dataValueChanged):
        dataWidget.setDataValue(MISSING)
    assert dataWidget.dataValue() == dict(a=MISSING, b=dict(x=False))


def test_DataWidget_subwidget(qtbot):
    @dataclasses.dataclass
    class Cls1:
        x: bool

    @dataclasses.dataclass
    class Cls2:
        a: int
        b: Cls1

    dataWidget = dataclass2Widget(Cls2)

    with qtbot.waitSignal(
        dataWidget.dataValueChanged,
        check_params_cb=lambda val: val == dict(a=MISSING, b=dict(x=True)),
    ):
        dataWidget.widget(1).widget(0).click()
    assert dataWidget.dataValue() == dict(a=MISSING, b=dict(x=True))

    with qtbot.waitSignal(
        dataWidget.dataValueChanged,
        check_params_cb=lambda tup: tup == dict(a=1, b=dict(x=True)),
    ):
        qtbot.keyPress(dataWidget.widget(0), "1")
        qtbot.keyPress(dataWidget.widget(0), QtCore.Qt.Key.Key_Return)
    assert dataWidget.dataValue() == dict(a=1, b=dict(x=True))


def test_type2Widget(qtbot):
    assert isinstance(type2Widget(bool), BoolCheckBox)
    assert not type2Widget(bool).isTristate()
    assert isinstance(type2Widget(Optional[bool]), BoolCheckBox)
    assert type2Widget(Optional[bool]).isTristate()

    assert isinstance(type2Widget(int), IntLineEdit)

    assert isinstance(type2Widget(float), FloatLineEdit)

    assert isinstance(type2Widget(str), StrLineEdit)

    class E(Enum):
        x = 1

    assert isinstance(type2Widget(E), EnumComboBox)

    with pytest.raises(TypeError):
        type2Widget(Tuple)
    with pytest.raises(TypeError):
        type2Widget(Tuple[int, ...])
    assert isinstance(type2Widget(Tuple[int, bool, E]), TupleGroupBox)


def test_dataclass2Widget(qtbot):
    @dataclasses.dataclass
    class Cls1:
        x: int

    @dataclasses.dataclass
    class Cls2:
        a: "int"
        b: bool
        c: float = dataclasses.field(metadata=dict(Qt_typehint=int))
        d: Cls1

    dataWidget = dataclass2Widget(Cls2)

    assert isinstance(dataWidget.widget(0), IntLineEdit)
    assert dataWidget.widget(0).fieldName() == "a"

    assert isinstance(dataWidget.widget(1), BoolCheckBox)
    assert dataWidget.widget(1).fieldName() == "b"

    assert isinstance(dataWidget.widget(2), IntLineEdit)
    assert dataWidget.widget(2).fieldName() == "c"

    assert isinstance(dataWidget.widget(3), DataWidget)
    assert dataWidget.widget(3).fieldName() == "d"
    assert isinstance(dataWidget.widget(3).widget(0), IntLineEdit)
    assert dataWidget.widget(3).widget(0).fieldName() == "x"



"""
Data field widgets
==================

:mod:`dawiq.fieldwidgets` provides widgets to represent the fields of the
dataclass. Widgets are compatible to :class:`dawiq.typing.FieldWidgetProtocol`.
"""

from .qt_compat import QtCore, QtWidgets, QtGui
from enum import Enum
from typing import Optional, Union, Tuple, TypeVar, Type
from .typing import FieldWidgetProtocol


__all__ = [
    "MISSING",
    "BoolCheckBox",
    "EmptyIntValidator",
    "IntLineEdit",
    "EmptyFloatValidator",
    "FloatLineEdit",
    "StrLineEdit",
    "EnumComboBox",
    "TupleGroupBox",
]


class _MISSING:
    """Sentinel object to indicate empty field."""

    pass


MISSING = _MISSING()


class BoolCheckBox(QtWidgets.QCheckBox):
    """
    Checkbox for fuzzy boolean value.

    :meth:`dataValue` returns the current value. When the check state is changed
    by user, :attr:`dataValueChanged` signal is emitted. :meth:`setDataValue`
    changes the check state of the checkbox.

    If the box is checked, the data value is True. If unchecked, the value is
    False. Else, e.g. ``Qt.PartiallyChecked``, the value is None.

    Because of the nature of check box, it is impossible to define empty state of
    the widget. :meth:`dataValue` is always either True, False or None, and never
    :obj:`MISSING`. :meth:`setDataValue` treats :obj:`MISSING` as False.

    """

    dataValueChanged = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._block_dataValueChanged = False

        self.stateChanged.connect(self.emitDataValueChanged)

    def fieldName(self) -> str:
        return self.text()

    def setFieldName(self, name: str):
        self.setText(name)
        self.setToolTip(name)

    def dataValue(self) -> Optional[bool]:
        checkstate = self.checkState()
        if checkstate == QtCore.Qt.CheckState.Checked:
            state = True
        elif checkstate == QtCore.Qt.CheckState.Unchecked:
            state = False
        else:
            state = None
        return state

    def setDataValue(self, value: Union[Optional[bool], _MISSING]):
        if value is MISSING:
            value = False

        if value is True:
            state = QtCore.Qt.CheckState.Checked
        elif value is False:
            state = QtCore.Qt.CheckState.Unchecked
        else:
            state = QtCore.Qt.CheckState.PartiallyChecked

        self._block_dataValueChanged = True
        self.setCheckState(state)
        self._block_dataValueChanged = False

    def emitDataValueChanged(self, checkstate: Union[int, QtCore.Qt.CheckState]):
        if self._block_dataValueChanged:
            return

        checkstate = QtCore.Qt.CheckState(checkstate)
        if checkstate == QtCore.Qt.CheckState.Checked:
            state = True
        elif checkstate == QtCore.Qt.CheckState.Unchecked:
            state = False
        else:
            state = None
        self.dataValueChanged.emit(state)


class EmptyIntValidator(QtGui.QIntValidator):
    """Validator which accpets integer and empty string"""

    def validate(self, input: str, pos: int) -> Tuple[QtGui.QValidator.State, str, int]:
        state, ret_input, ret_pos = super().validate(input, pos)
        if not input:
            state = QtGui.QValidator.State.Acceptable
        return (state, ret_input, ret_pos)


class IntLineEdit(QtWidgets.QLineEdit):
    """
    Line edit for integer value.

    :meth:`dataValue` returns the current value. When editing is finished,
    :attr:`dataValueChanged` signal is emitted. :meth:`setDataValue` changes the
    text on the line edit.

    If the text is not empty, the data value is the integer that the string is
    converted to. If the line edit is empty, :obj:`MISSING` is the data value.

    :meth:`setDataValue` sets the line edit text. If :obj:`MISSING` is passed,
    line edit is cleared.

    """

    dataValueChanged = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setValidator(EmptyIntValidator(self))

        self.editingFinished.connect(self.emitDataValueChanged)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def dataValue(self) -> Union[int, _MISSING]:
        text = self.text()

        if not text:
            val: Union[int, _MISSING] = MISSING
        else:
            try:
                val = int(text)
            except ValueError:
                val = MISSING
        return val

    def setDataValue(self, val: Union[int, _MISSING]):
        if val is MISSING:
            txt = ""
        else:
            txt = str(val)
        self.setText(txt)

    def emitDataValueChanged(self):
        val = self.dataValue()
        self.dataValueChanged.emit(val)


class EmptyFloatValidator(QtGui.QDoubleValidator):
    """Validator which accpets float and empty string"""

    def validate(self, input: str, pos: int) -> Tuple[QtGui.QValidator.State, str, int]:
        state, ret_input, ret_pos = super().validate(input, pos)
        if not input:
            state = QtGui.QValidator.State.Acceptable
        return (state, ret_input, ret_pos)


class FloatLineEdit(QtWidgets.QLineEdit):
    """
    Line edit for float value.

    :meth:`dataValue` returns the current value. When editing is finished,
    :attr:`dataValueChanged` signal is emitted. :meth:`setDataValue` changes the
    text on the line edit.

    If the text is not empty, the data value is the float that the string is
    converted to. If the line edit is empty, :obj:`MISSING` is the data value.

    :meth:`setDataValue` sets the line edit text. If :obj:`MISSING` is passed,
    line edit is cleared.

    """

    dataValueChanged = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setValidator(EmptyFloatValidator(self))

        self.editingFinished.connect(self.emitDataValueChanged)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def dataValue(self) -> Union[float, _MISSING]:
        text = self.text()

        if not text:
            val: Union[float, _MISSING] = MISSING
        else:
            try:
                val = float(text)
            except ValueError:
                val = MISSING
        return val

    def setDataValue(self, val: Union[float, _MISSING]):
        if val is MISSING:
            txt = ""
        else:
            txt = str(val)
        self.setText(txt)

    def emitDataValueChanged(self):
        val = self.dataValue()
        self.dataValueChanged.emit(val)


class StrLineEdit(QtWidgets.QLineEdit):
    """
    Line edit for string value.

    :meth:`dataValue` returns the current value. When editing is finished,
    :attr:`dataValueChanged` signal is emitted. :meth:`setDataValue` changes the
    text on the line edit.

    Data value is the text of line edit. If the line edit is empty, data value is
    empty string. Thus, the data value is never :obj:`MISSING`.

    :meth:`setDataValue` sets the line edit text. If :obj:`MISSING` is passed,
    line edit is cleared.

    """

    dataValueChanged = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.editingFinished.connect(self.emitDataValueChanged)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def dataValue(self) -> str:
        return self.text()

    def setDataValue(self, val: Union[str, _MISSING]):
        if val is MISSING:
            txt = ""
        else:
            txt = val  # type: ignore[assignment]
        self.setText(txt)

    def emitDataValueChanged(self):
        val = self.dataValue()
        self.dataValueChanged.emit(val)


T = TypeVar("T", bound="EnumComboBox")


class EnumComboBox(QtWidgets.QComboBox):
    """
    Combo box for enum type.

    Standard way to construct this widget is by :meth:`fromEnum` class method.
    N-th item contains N-th member of the Enum as its data.

    :meth:`dataValue` returns the current member. When current index is changed
    by user, :attr:`dataValueChanged` signal is emitted. :meth:`setDataValue`
    changes the current index.

    Data value is the Enum member. If the current index is empty, data value is
    :obj:`MISSING`

    :meth:`setDataValue` sets the current index. If :obj:`MISSING` is passed,
    index is set to -1.

    """

    dataValueChanged = QtCore.Signal(object)

    @classmethod
    def fromEnum(cls: Type[T], enum: Type[Enum]) -> T:
        obj = cls()
        for e in enum:
            obj.addItem(e.name, userData=e)
        obj.setCurrentIndex(-1)
        return obj

    def __init__(self, parent=None):
        super().__init__(parent)
        self._block_dataValueChanged = False

        self.currentIndexChanged.connect(self.emitDataValueChanged)

    def fieldName(self) -> str:
        return self.placeholderText()

    def setFieldName(self, name: str):
        self.setPlaceholderText(name)
        self.setToolTip(name)

    def dataValue(self) -> Union[Enum, _MISSING]:
        index = self.currentIndex()
        if index == -1:
            ret = MISSING
        else:
            ret = self.itemData(index)
        return ret

    def setDataValue(self, value: Union[Enum, _MISSING]):
        if value is MISSING:
            index = -1
        else:
            index = self.findData(value)
        self._block_dataValueChanged = True
        self.setCurrentIndex(index)
        self._block_dataValueChanged = False

    def emitDataValueChanged(self):
        if self._block_dataValueChanged:
            return
        val = self.dataValue()
        self.dataValueChanged.emit(val)


V = TypeVar("V", bound="TupleGroupBox")


class TupleGroupBox(QtWidgets.QGroupBox):
    """
    Group box for tuple data with fixed length.

    This is the group box which contains field widgets as subwidgets. Data value
    is constructed from the data of subwidgets as tuple.

    :meth:`dataValue` returns the current tuple value. When data value of any
    subwidget is changed by user, :attr:`dataValueChanged` signal is emitted.
    :meth:`setDataValue` changes the data of subwidgets.

    Data value is the tuple containing subwidget data, and never :obj:`MISSING`.

    :meth:`setDataValue` sets the subwidget data. If :obj:`MISSING` is passed,
    it is propagated to all subwidget.

    """

    dataValueChanged = QtCore.Signal(tuple)

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation = QtCore.Qt.Orientation.Horizontal,
        parent=None,
    ):
        super().__init__(parent)
        self._orientation = orientation
        self._block_dataValueChanged = False

        if orientation == QtCore.Qt.Orientation.Vertical:
            layout = QtWidgets.QVBoxLayout()
        elif orientation == QtCore.Qt.Orientation.Horizontal:
            layout = QtWidgets.QHBoxLayout()
        else:
            raise TypeError(f"Invalid orientation: {orientation}")
        self.setLayout(layout)

    def fieldName(self) -> str:
        return self.title()

    def setFieldName(self, name: str):
        self.setTitle(name)
        self.setToolTip(name)

    def orientation(self) -> QtCore.Qt.Orientation:
        """Orientation to stack the subwidgets."""
        return self._orientation

    def count(self) -> int:
        """Number of subwidgets."""
        return self.layout().count()

    def widget(self, index: int) -> Optional[FieldWidgetProtocol]:
        """
        Returns the subwidget at the given index, or None for invalid index.
        """
        item = self.layout().itemAt(index)
        if item is not None:
            item = item.widget()
        return item

    def insertWidget(
        self,
        index: int,
        widget: FieldWidgetProtocol,
        stretch: int = 0,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag(0),
    ):
        """Insert the widget to layout and connect data value change signal."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
        widget.dataValueChanged.connect(self.emitDataValueChanged)
        self.layout().insertWidget(index, widget, stretch, alignment)

    def addWidget(
        self,
        widget: FieldWidgetProtocol,
        stretch: int = 0,
        alignment: QtCore.Qt.AlignmentFlag = QtCore.Qt.AlignmentFlag(0),
    ):
        """Add the widget to layout and connect data value change signal."""
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
        widget.dataValueChanged.connect(self.emitDataValueChanged)
        self.layout().addWidget(widget, stretch, alignment)

    def removeWidget(self, widget: FieldWidgetProtocol):
        """
        Remove the widget from layout and disconnect data value change signal.
        """
        for i in range(self.count()):
            w = self.widget(i)
            if w is None:
                break
            if w == widget:
                widget.dataValueChanged.disconnect(self.emitDataValueChanged)
                break
        self.layout().removeWidget(widget)

    def dataValue(self) -> tuple:
        ret = []
        for i in range(self.count()):
            widget = self.widget(i)
            if widget is None:
                break
            ret.append(widget.dataValue())
        return tuple(ret)

    def setDataValue(self, value: Union[tuple, _MISSING]):
        self._block_dataValueChanged = True
        if value is MISSING:
            for i in range(self.count()):
                widget = self.widget(i)
                if widget is None:
                    break
                widget.setDataValue(MISSING)
        else:
            for i in range(self.count()):
                widget = self.widget(i)
                if widget is None:
                    break
                widget.setDataValue(value[i])  # type: ignore[index]
        self._block_dataValueChanged = False

    def emitDataValueChanged(self):
        if self._block_dataValueChanged:
            return
        val = self.dataValue()
        self.dataValueChanged.emit(val)



"""
Dataclass delegate
==================

"""

import dataclasses
from .qt_compat import QtWidgets
from .fieldwidgets import MISSING
from .datawidget import DataWidget
from .typing import DataclassProtocol
from typing import Optional, Type, Dict, Any


__all__ = [
    "convertFromQt",
    "convertToQt",
    "DataclassDelegate",
    "DataclassMapper",
]


def convertFromQt(
    dcls: Type[DataclassProtocol],
    data: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Convert dict from :class:`DataWidget` to structured dict for dataclass.

    If the field value does not exist or is :obj:`MISSING`, the field is not
    included in the resulting dictionary. Default value of the field is ignored.

    Field may define `fromQt_converter` metadata to convert the widget data to
    field data. It is a unary callable which takes the widget data and returns
    the field data.

    Examples
    ========

    >>> from dataclasses import dataclass, field
    >>> from dawiq import MISSING
    >>> from dawiq.delegate import convertFromQt
    >>> def conv(arg):
    ...     return (arg,)
    >>> @dataclass
    ... class Cls:
    ...     x: tuple = field(metadata=dict(fromQt_converter=conv), default=(1,))
    >>> convertFromQt(Cls, dict(x=10))
    {'x': (10,)}
    >>> convertFromQt(Cls, dict(x=MISSING))
    {}

    """
    # Return value is not dataclass but dictionary because necessary fields might
    # be missing from the widget.
    ret = {}
    for f in dataclasses.fields(dcls):
        val = data.get(f.name, MISSING)
        if val is MISSING:
            continue

        if dataclasses.is_dataclass(f.type):
            val = convertFromQt(f.type, val)

        converter = f.metadata.get("fromQt_converter", None)
        if converter is not None:
            val = converter(val)
        ret[f.name] = val
    return ret


def convertToQt(
    dcls: Type[DataclassProtocol],
    data: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Convert structured dict from dataclass to dict for :class:`DataWidget`.

    If the data does not have the value for a field, :obj:`MISSING` is passed as
    its value instead.

    Field may define `toQt_converter` metadata to convert the field data to
    widget data. It is a unary callable which takes the field data and returns
    the widget data.

    Examples
    ========

    >>> from dataclasses import dataclass, field
    >>> from dawiq import MISSING
    >>> from dawiq.delegate import convertToQt
    >>> @dataclass
    ... class Cls:
    ...     x: tuple = field(metadata=dict(toQt_converter=lambda tup: tup[0]))
    >>> convertToQt(Cls, dict(x=(10,)))
    {'x': 10}
    >>> convertToQt(Cls, dict()) == dict(x=MISSING)
    True

    """
    ret = {}
    for f in dataclasses.fields(dcls):
        val = data.get(f.name, MISSING)
        if val is MISSING:
            pass
        elif dataclasses.is_dataclass(f.type):
            val = convertToQt(f.type, val)
        converter = f.metadata.get("toQt_converter", None)
        if val is not MISSING and converter is not None:
            val = converter(val)
        ret[f.name] = val
    return ret


class DataclassDelegate(QtWidgets.QStyledItemDelegate):
    """
    Delegate to update the model and editor with structured dictionary.

    By setting :meth:`dataclassType`, this delegate can convert the widget data
    to field data and vice versa.

    Even if the fields of :meth:`dataclassType` has default value, it is not
    applied to the widget and the model. This is to make sure that empty input
    by user is distinguished.

    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._dataclass_type = None

    def dataclassType(self) -> Optional[Type[DataclassProtocol]]:
        """
        Dataclass type whose fields are used to convert the data.

        ``None`` indicates that the dataclass type is not set, and data will
        not be converted.
        """
        return self._dataclass_type

    def setDataclassType(self, dcls: Optional[Type[DataclassProtocol]]):
        self._dataclass_type = dcls

    def setModelData(self, editor, model, index):
        """
        Set the data from *editor* to the item of *model* at *index*.

        If *editor* is :class:`DataWidget` and :meth:`dataclassType` is set, its
        data is converted by :func:`convertFromQt` before being set to the model.
        """
        if isinstance(editor, DataWidget):
            dcls = self.dataclassType()
            data = editor.dataValue()
            if dcls is not None:
                data = convertFromQt(dcls, data)
            model.setData(index, data)
        else:
            super().setModelData(editor, model, index)

    def setEditorData(self, editor, index):
        """
        Set the data from *index* to *editor*.

        If *editor* is :class:`DataWidget` and :meth:`dataclassType` is set, the
        data is converted by :func:`convertToQt` before being set to the editor.
        """
        if isinstance(editor, DataWidget):
            dcls = self.dataclassType()
            data = index.data()
            if data is None:
                data = {}
            if dcls is not None:
                data = convertToQt(dcls, data)
            editor.setDataValue(data)
        else:
            super().setEditorData(editor, index)


class DataclassMapper(QtWidgets.QDataWidgetMapper):
    """
    Mapper between :class:`DataWidget` and model.

    Notes
    =====

    When mapping :class:`DataWidget`, *propertyName* argument of
    :meth:`addMapping` must not be passed.
    """

    def addMapping(self, widget, section, propertyName=b""):
        super().addMapping(widget, section, propertyName)
        if isinstance(widget, DataWidget):
            widget.dataValueChanged.connect(self.submit)

    def removeMapping(self, widget):
        super().removeMapping(widget)
        if isinstance(widget, DataWidget):
            widget.dataValueChanged.disconnect(self.submit)



import dataclasses
from dawiq import dataclass2Widget, MISSING
from dawiq.delegate import (
    convertFromQt,
    convertToQt,
    DataclassDelegate,
    DataclassMapper,
)
from dawiq.qt_compat import QtGui, QtWidgets, QtCore
from typing import Tuple


def test_convertFromQt():
    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        a: CustomField = dataclasses.field(
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg))
        )

    @dataclasses.dataclass
    class Cls1:
        x: int
        y: CustomField = dataclasses.field(
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg))
        )
        z: Cls0

    assert convertFromQt(Cls1, dict(x=1, y=2, z=dict(a=3))) == dict(
        x=1, y=CustomField(2), z=dict(a=CustomField(3))
    )
    assert convertFromQt(Cls1, dict(x=MISSING, y=MISSING, z=MISSING)) == dict()


def test_convertFromQt_defaultvalue():
    """Test that default value is ignored."""

    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        x: CustomField = dataclasses.field(
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg))
        )
        y: CustomField = dataclasses.field(
            default=CustomField(0),
            metadata=dict(fromQt_converter=lambda arg: CustomField(arg)),
        )
        z: int = 3

    assert convertFromQt(Cls0, dict(x=3, y=2, z=1)) == dict(
        x=CustomField(3), y=CustomField(2), z=1
    )
    assert convertFromQt(Cls0, dict()) == dict()
    assert convertFromQt(Cls0, dict(x=MISSING, y=MISSING, z=MISSING)) == dict()

    @dataclasses.dataclass
    class Cls1:
        a: Cls0
        b: Cls0 = Cls0(x=CustomField(1))

    assert convertFromQt(
        Cls1, dict(a=dict(x=1, y=2, z=5), b=dict(x=3, y=2, z=1))
    ) == dict(
        a=dict(x=CustomField(1), y=CustomField(2), z=5),
        b=dict(x=CustomField(3), y=CustomField(2), z=1),
    )
    assert convertFromQt(Cls1, dict()) == dict()
    assert convertFromQt(Cls1, dict(a=MISSING, b=MISSING)) == dict()
    assert convertFromQt(
        Cls1, dict(a=MISSING, b=dict(x=MISSING, y=MISSING, z=MISSING))
    ) == dict(b=dict())

    @dataclasses.dataclass
    class Cls2:
        c: Cls1
        d: Cls1 = Cls1(Cls0(CustomField(10)))

    assert convertFromQt(Cls2, dict()) == dict()


def test_convertToQt():
    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        a: CustomField = dataclasses.field(
            metadata=dict(toQt_converter=lambda val: val.x)
        )

    @dataclasses.dataclass
    class Cls1:
        x: int
        y: CustomField = dataclasses.field(
            metadata=dict(toQt_converter=lambda val: val.x)
        )
        z: Cls0

    assert convertToQt(
        Cls1, dict(x=1, y=CustomField(2), z=dict(a=CustomField(3)))
    ) == dict(x=1, y=2, z=dict(a=3))
    assert convertToQt(Cls1, dict()) == dict(x=MISSING, y=MISSING, z=MISSING)


def test_convertToQt_defaultvalue():
    """Test that default value is ignored."""

    class CustomField:
        def __init__(self, x):
            self.x = x

        def __eq__(self, other):
            return type(self) == type(other) and self.x == other.x

    @dataclasses.dataclass
    class Cls0:
        x: CustomField = dataclasses.field(
            metadata=dict(toQt_converter=lambda val: val.x)
        )
        y: CustomField = dataclasses.field(
            default=CustomField(0),
            metadata=dict(toQt_converter=lambda val: val.x),
        )
        z: int = 3

    assert convertToQt(Cls0, dict()) == dict(x=MISSING, y=MISSING, z=MISSING)

    @dataclasses.dataclass
    class Cls1:
        a: Cls0
        b: Cls0 = Cls0(x=CustomField(1))

    assert convertToQt(Cls1, dict()) == dict(a=MISSING, b=MISSING)
    assert convertToQt(Cls1, dict(b=dict())) == dict(
        a=MISSING, b=dict(x=MISSING, y=MISSING, z=MISSING)
    )

    @dataclasses.dataclass
    class Cls2:
        c: Cls1
        d: Cls1 = Cls1(Cls0(CustomField(10)))

    assert convertToQt(Cls2, dict()) == dict(c=MISSING, d=MISSING)
    assert convertToQt(Cls2, dict(d=dict())) == dict(
        c=MISSING, d=dict(a=MISSING, b=MISSING)
    )


def test_DataclassDelegate_setModelData(qtbot):
    model = QtGui.QStandardItemModel()
    model.appendRow(QtGui.QStandardItem())

    @dataclasses.dataclass
    class Dcls:
        x: int

    dataWidget = dataclass2Widget(Dcls)
    mapper = QtWidgets.QDataWidgetMapper()
    delegate = DataclassDelegate()
    delegate.setDataclassType(Dcls)

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex) is None

    delegate.commitData.emit(dataWidget)
    assert model.data(modelIndex) == dict()

    dataWidget.widget(0).setText("0")
    delegate.commitData.emit(dataWidget)
    assert model.data(modelIndex) == dict(x=0)

    dataWidget.widget(0).setText("1")
    mapper.submit()
    assert model.data(modelIndex) == dict(x=1)

    dataWidget.dataValueChanged.connect(mapper.submit)
    with qtbot.waitSignal(dataWidget.dataValueChanged):
        dataWidget.widget(0).setText("2")
        qtbot.keyPress(dataWidget.widget(0), QtCore.Qt.Key.Key_Return)
    assert model.data(modelIndex) == dict(x=2)


def test_DataclassDelegate_setEditorData(qtbot):
    model = QtGui.QStandardItemModel()
    for i in range(3):
        model.appendRow(QtGui.QStandardItem())

    @dataclasses.dataclass
    class Dcls:
        x: int

    dataWidget = dataclass2Widget(Dcls)
    mapper = QtWidgets.QDataWidgetMapper()
    delegate = DataclassDelegate()
    delegate.setDataclassType(Dcls)

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex0 = model.index(0, 0)
    model.setData(modelIndex0, dict(x=0))
    modelIndex1 = model.index(1, 0)
    model.setData(modelIndex1, dict(x=1))
    modelIndex2 = model.index(2, 0)
    model.setData(modelIndex2, dict())

    assert dataWidget.dataValue() == dict(x=MISSING)
    assert dataWidget.widget(0).text() == ""

    mapper.setCurrentModelIndex(modelIndex0)
    assert dataWidget.dataValue() == dict(x=0)
    assert dataWidget.widget(0).text() == "0"

    mapper.setCurrentModelIndex(modelIndex1)
    assert dataWidget.dataValue() == dict(x=1)
    assert dataWidget.widget(0).text() == "1"

    mapper.setCurrentModelIndex(modelIndex2)
    assert dataWidget.dataValue() == dict(x=MISSING)
    assert dataWidget.widget(0).text() == ""

    model.setData(modelIndex2, dict(x=10))
    assert dataWidget.dataValue() == dict(x=10)
    assert dataWidget.widget(0).text() == "10"


def test_DataclassMapper_addMapping(qtbot):
    model = QtGui.QStandardItemModel()
    model.appendRow(QtGui.QStandardItem())

    @dataclasses.dataclass
    class Dcls:
        x: int
        y: bool

    dataWidget = dataclass2Widget(Dcls)
    mapper = DataclassMapper()
    delegate = DataclassDelegate()
    delegate.setDataclassType(Dcls)

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    # assert model.data(modelIndex) == dict(y=False)

    dataWidget.widget(0).setText("0")
    qtbot.keyPress(dataWidget.widget(0), QtCore.Qt.Key.Key_Return)
    assert model.data(modelIndex) == dict(x=0, y=False)

    dataWidget.widget(1).click()
    assert model.data(modelIndex) == dict(x=0, y=True)


def test_DataclassMapper_removeMapping(qtbot):
    @dataclasses.dataclass
    class Dcls:
        x: int
        y: bool

    dataWidget = dataclass2Widget(Dcls)
    model = QtGui.QStandardItemModel()
    model.appendRow(QtGui.QStandardItem())
    mapper = DataclassMapper()
    delegate = DataclassDelegate()
    delegate.setDataclassType(Dcls)

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    mapper.removeMapping(dataWidget)

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex) is None

    dataWidget.widget(0).setText("0")
    qtbot.keyPress(dataWidget.widget(0), QtCore.Qt.Key.Key_Return)
    assert model.data(modelIndex) is None

    dataWidget.widget(1).click()
    assert model.data(modelIndex) is None


def test_DataclassMapper_clearMapping(qtbot):
    @dataclasses.dataclass
    class Dcls:
        x: int
        y: bool

    dataWidget = dataclass2Widget(Dcls)
    model = QtGui.QStandardItemModel()
    model.appendRow(QtGui.QStandardItem())
    mapper = DataclassMapper()
    delegate = DataclassDelegate()
    delegate.setDataclassType(Dcls)

    mapper.setModel(model)
    mapper.addMapping(dataWidget, 0)
    mapper.setItemDelegate(delegate)

    mapper.clearMapping()

    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)
    assert model.data(modelIndex) is None

    dataWidget.widget(0).setText("0")
    qtbot.keyPress(dataWidget.widget(0), QtCore.Qt.Key.Key_Return)
    assert model.data(modelIndex) is None

    dataWidget.widget(1).click()
    assert model.data(modelIndex) is None


def test_DataclassMapper_Tuple_setCurrentIndex_crash(qtbot):
    """Test that setting index to nested widget does not cause infinite loop."""

    @dataclasses.dataclass
    class DataClass:
        x: Tuple[int]

    delegate = DataclassDelegate()
    delegate.setDataclassType(DataClass)
    mapper = DataclassMapper()
    mapper.setItemDelegate(delegate)

    model = QtGui.QStandardItemModel()
    model.appendRow(QtGui.QStandardItem())
    mapper.setModel(model)

    dataWidget = dataclass2Widget(DataClass)
    mapper.addMapping(dataWidget, 0)
    mapper.setCurrentIndex(0)  # must not crash


def test_DataclassMapper_default(qtbot):
    """Test that default value of dataclass is not applied to widget & model."""

    @dataclasses.dataclass
    class DataClass:
        x: int = 3

    delegate = DataclassDelegate()
    delegate.setDataclassType(DataClass)
    mapper = DataclassMapper()
    mapper.setItemDelegate(delegate)

    model = QtGui.QStandardItemModel()
    model.appendRow(QtGui.QStandardItem())
    mapper.setModel(model)

    dataWidget = dataclass2Widget(DataClass)
    mapper.addMapping(dataWidget, 0)
    modelIndex = model.index(0, 0)
    mapper.setCurrentModelIndex(modelIndex)

    assert model.data(modelIndex) is None
    assert dataWidget.dataValue() == dict(x=MISSING)



from dawiq import (
    BoolCheckBox,
    EmptyIntValidator,
    IntLineEdit,
    EmptyFloatValidator,
    FloatLineEdit,
    StrLineEdit,
    EnumComboBox,
    TupleGroupBox,
    MISSING,
)
import enum
from dawiq.qt_compat import QtCore, QtWidgets


def test_BoolCheckBox(qtbot):
    widget = BoolCheckBox()
    widget.setTristate(True)
    widget.setCheckState(QtCore.Qt.CheckState.Unchecked)

    # test value change by setDataValue
    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(True)
    assert widget.dataValue() is True
    assert widget.checkState() == QtCore.Qt.CheckState.Checked

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(False)
    assert widget.dataValue() is False
    assert widget.checkState() == QtCore.Qt.CheckState.Unchecked

    widget.setCheckState(QtCore.Qt.CheckState.Checked)  # set to True to test MISSING
    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MISSING)
    assert widget.dataValue() is False
    assert widget.checkState() == QtCore.Qt.CheckState.Unchecked

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(None)
    assert widget.dataValue() is None
    assert widget.checkState() == QtCore.Qt.CheckState.PartiallyChecked

    widget.setCheckState(QtCore.Qt.CheckState.Unchecked)

    # test value change by clicking
    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val is None,
    ):
        widget.click()
    assert widget.dataValue() is None
    assert widget.checkState() == QtCore.Qt.CheckState.PartiallyChecked

    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val is True,
    ):
        widget.click()
    assert widget.dataValue() is True
    assert widget.checkState() == QtCore.Qt.CheckState.Checked

    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val is False,
    ):
        widget.click()
    assert widget.dataValue() is False
    assert widget.checkState() == QtCore.Qt.CheckState.Unchecked


def test_EmptyIntValidator(qtbot):
    widget = QtWidgets.QLineEdit()
    widget.setValidator(EmptyIntValidator(widget))

    # empty text is valid
    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.assertNotEmitted(widget.editingFinished):
        qtbot.keyPress(widget, "-")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)


def test_IntLineEdit(qtbot):
    widget = IntLineEdit()

    # test value change by setDataValue
    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MISSING)
    assert widget.dataValue() is MISSING
    assert not widget.text()

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(1)
    assert widget.dataValue() == 1
    assert widget.text() == "1"

    widget.clear()

    # test value change by keyboard
    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val is MISSING,
    ):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() is MISSING

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        qtbot.keyPress(widget, "-")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() is MISSING

    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val == -1,
    ):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() == -1


def test_EmptyFloatValidator(qtbot):
    widget = QtWidgets.QLineEdit()
    widget.setValidator(EmptyFloatValidator(widget))

    # empty text is valid
    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.assertNotEmitted(widget.editingFinished):
        qtbot.keyPress(widget, "-")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, ".")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)

    with qtbot.waitSignal(widget.editingFinished):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)


def test_FloatLineEdit(qtbot):
    widget = FloatLineEdit()

    # test value change by setDataValue
    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MISSING)
    assert widget.dataValue() is MISSING
    assert not widget.text()

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(1)
    assert widget.dataValue() == float(1)
    assert widget.text() == "1"

    widget.clear()

    # test value change by keyboard
    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val is MISSING,
    ):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() is MISSING

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        qtbot.keyPress(widget, "-")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() is MISSING

    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val == float(-1),
    ):
        qtbot.keyPress(widget, "1")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() == float(-1)


def test_StrLineEdit(qtbot):
    widget = StrLineEdit()

    # test value change by setDataValue
    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MISSING)
    assert widget.dataValue() == ""
    assert not widget.text()

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue("1")
    assert widget.dataValue() == "1"
    assert widget.text() == "1"

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue("x")
    assert widget.dataValue() == "x"
    assert widget.text() == "x"

    widget.clear()

    # test value change by keyboard
    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val == "",
    ):
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() == ""

    with qtbot.waitSignal(
        widget.dataValueChanged,
        check_params_cb=lambda val: val == "x",
    ):
        qtbot.keyPress(widget, "x")
        qtbot.keyPress(widget, QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() == "x"


def test_EnumComboBox(qtbot):
    class MyEnum(enum.Enum):
        x = 1
        y = 2
        z = 3

    widget = EnumComboBox.fromEnum(MyEnum)
    assert widget.count() == 3
    assert widget.currentIndex() == -1
    assert widget.dataValue() is MISSING

    # test with setDataValue
    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MyEnum.x)
    assert widget.currentIndex() == 0
    assert widget.dataValue() == MyEnum.x

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MyEnum.y)
    assert widget.currentIndex() == 1
    assert widget.dataValue() == MyEnum.y

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MyEnum.z)
    assert widget.currentIndex() == 2
    assert widget.dataValue() == MyEnum.z

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MISSING)
    assert widget.currentIndex() == -1
    assert widget.dataValue() is MISSING

    # test with setCurrentIndex

    with qtbot.waitSignal(
        widget.dataValueChanged,
        raising=True,
        check_params_cb=lambda val: val == MyEnum.x,
    ):
        widget.setCurrentIndex(0)
    assert widget.dataValue() == MyEnum.x

    with qtbot.waitSignal(
        widget.dataValueChanged,
        raising=True,
        check_params_cb=lambda val: val == MyEnum.y,
    ):
        widget.setCurrentIndex(1)
    assert widget.dataValue() == MyEnum.y

    with qtbot.waitSignal(
        widget.dataValueChanged,
        raising=True,
        check_params_cb=lambda val: val == MyEnum.z,
    ):
        widget.setCurrentIndex(2)
    assert widget.dataValue() == MyEnum.z

    with qtbot.waitSignal(
        widget.dataValueChanged,
        raising=True,
        check_params_cb=lambda val: val is MISSING,
    ):
        widget.setCurrentIndex(-1)
    assert widget.dataValue() is MISSING


def test_TupleGroupBox_addWidget(qtbot):
    tupleWidget = TupleGroupBox()
    assert tupleWidget.count() == 0

    w0 = BoolCheckBox()
    tupleWidget.addWidget(w0)
    assert tupleWidget.count() == 1
    assert tupleWidget.widget(0) is w0
    assert tupleWidget.widget(1) is None

    w1 = IntLineEdit()
    tupleWidget.addWidget(w1)
    assert tupleWidget.count() == 2
    assert tupleWidget.widget(0) is w0
    assert tupleWidget.widget(1) is w1

    # test that signals are connected
    with qtbot.waitSignal(tupleWidget.dataValueChanged):
        w0.click()
    with qtbot.waitSignal(tupleWidget.dataValueChanged):
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)


def test_TupleGroupBox_insertWidget(qtbot):
    tupleWidget = TupleGroupBox()
    assert tupleWidget.count() == 0

    w0 = BoolCheckBox()
    tupleWidget.insertWidget(0, w0)
    assert tupleWidget.count() == 1
    assert tupleWidget.widget(0) is w0
    assert tupleWidget.widget(1) is None

    w1 = IntLineEdit()
    tupleWidget.insertWidget(0, w1)
    assert tupleWidget.count() == 2
    assert tupleWidget.widget(0) is w1
    assert tupleWidget.widget(1) is w0

    # test that signals are connected
    with qtbot.waitSignal(tupleWidget.dataValueChanged):
        w0.click()
    with qtbot.waitSignal(tupleWidget.dataValueChanged):
        qtbot.keyPress(w1, QtCore.Qt.Key.Key_Return)


def test_TupleGroupBox_removeWidget(qtbot):
    tupleWidget = TupleGroupBox()
    w0 = BoolCheckBox()
    w1 = IntLineEdit()

    tupleWidget.addWidget(w0)
    assert tupleWidget.count() == 1

    tupleWidget.removeWidget(w1)
    assert tupleWidget.count() == 1

    tupleWidget.removeWidget(w0)
    assert tupleWidget.count() == 0

    # test that signals are disconnected
    with qtbot.assertNotEmitted(tupleWidget.dataValueChanged):
        w0.click()


def test_TupleGroupBox_dataValue(qtbot):
    widget = TupleGroupBox()
    widget.addWidget(IntLineEdit())
    widget.addWidget(IntLineEdit())
    assert widget.dataValue() == (MISSING, MISSING)

    widget.widget(0).setText("1")
    widget.widget(1).setText("2")
    assert widget.dataValue() == (1, 2)


def test_TupleGroupBox_setDataValue(qtbot):
    widget = TupleGroupBox()
    widget.addWidget(IntLineEdit())
    widget.addWidget(IntLineEdit())

    class Counter:
        def __init__(self):
            self.i = 0

        def count(self):
            self.i += 1

    counter = Counter()
    widget.dataValueChanged.connect(counter.count)

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue((1, 2))
    assert widget.dataValue() == (1, 2)
    assert counter.i == 0

    with qtbot.assertNotEmitted(widget.dataValueChanged):
        widget.setDataValue(MISSING)
    assert widget.dataValue() == (MISSING, MISSING)


def test_TupleGroupBox_subwidget(qtbot):
    widget = TupleGroupBox()
    widget.addWidget(IntLineEdit())
    widget.addWidget(IntLineEdit())

    with qtbot.waitSignal(
        widget.dataValueChanged, check_params_cb=lambda tup: tup == (1, MISSING)
    ):
        qtbot.keyPress(widget.widget(0), "1")
        qtbot.keyPress(widget.widget(0), QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() == (1, MISSING)

    with qtbot.waitSignal(
        widget.dataValueChanged, check_params_cb=lambda tup: tup == (1, 2)
    ):
        qtbot.keyPress(widget.widget(1), "2")
        qtbot.keyPress(widget.widget(1), QtCore.Qt.Key.Key_Return)
    assert widget.dataValue() == (1, 2)
